///////////////////////////////////////////////////////////////////////////////
//
// Description
//
//  A string class.
//
// Edit History
//
//  Written October 1993 by Andrew Davison.
//
//  Made a bit safer, October 1995 by AD.
//  Use DLE escaping for ' ' & '\n' chararcters, November 1995 by AD.
//  Bring more in line with ANSI 'string' class, April 1996 by AD.
//
///////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <strstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "astring.h"
#pragma hdrstop

const DLE = 0x10;

AString AString::nil("");

AString::AString(char s)
{
	sref = new AStringRef;
	sref->refcnt = 1;
	sref->buf = new char[1+1];
	sref->nbytes = 1;
	sref->buf[0] = s;
	sref->buf[1] = 0;
}

AString::AString(const char* _s)
{
	const char* s = (_s ? _s : nil.c_str());

	unsigned len = (unsigned)strlen(s);
	sref = new AStringRef;
	sref->refcnt = 1;
	sref->buf = new char[len+1];
	sref->nbytes = len;
	memcpy(sref->buf, s, len);
	sref->buf[len] = 0;
}

AString::AString(const char* _s, unsigned len)
{
	const char* s = (_s ? _s : nil.c_str());

	unsigned newlen = (unsigned)strlen(s);

	if (newlen > len)
		newlen = len;

	sref = new AStringRef;
	sref->refcnt = 1;
	sref->buf = new char[newlen+1];
	sref->nbytes = newlen;
	memcpy(sref->buf, s, newlen);
	sref->buf[newlen] = 0;
}

AString::AString(const AString& s, unsigned offset, unsigned len)
{
	unsigned newlen = s.length();

	if (offset > newlen)
		offset = newlen;

	newlen -= offset;

	if (newlen > len)
		newlen = len;

	sref = new AStringRef;
	sref->refcnt = 1;
	sref->buf = new char[newlen+1];
	sref->nbytes = newlen;
	if (s.sref) memcpy(sref->buf, s.sref->buf+offset, newlen);
	sref->buf[newlen] = 0;
}

AString& AString::operator=(char s)
{
	AStringRef* old_sref = sref;

	sref = new AStringRef;
	sref->buf = new char[1+1];
	sref->refcnt = 1;
	sref->nbytes = 1;

	sref->buf[0] = s;
	sref->buf[1] = 0;

	if (old_sref)
	{
		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}

	return *this;
}

AString& AString::operator=(const char* _s)
{
	const char *s = (_s ? _s : nil.c_str());

	AStringRef* old_sref = sref;

	unsigned len = (unsigned)strlen(s);
	sref = new AStringRef;
	sref->buf = new char[len+1];
	sref->refcnt = 1;
	sref->nbytes = len;
	memcpy(sref->buf, s, len);
	sref->buf[len] = 0;

	if (old_sref)
	{
		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}

	return *this;
}

AString& AString::operator=(const AString& s)
{
	if (sref == s.sref)
		return *this;

	AStringRef* old_sref = sref;

	sref = s.sref;
	if (sref) sref->refcnt++;

	if (old_sref)
	{
		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}

	return *this;
}

AString& AString::operator+=(const AString& s)
{
	AStringRef* old_sref = sref;
	unsigned len = (s.sref ? s.sref->nbytes : (unsigned)0);

	sref = new AStringRef;
	sref->refcnt = 1;

	if (old_sref)
	{
		sref->buf = new char[old_sref->nbytes+len+1];
		sref->nbytes = old_sref->nbytes;
		memcpy(sref->buf, old_sref->buf, old_sref->nbytes);

		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}
	else
	{
		sref->buf = new char[len+1];
		sref->nbytes = 0;
	}

	if (s.sref) memcpy(sref->buf+sref->nbytes, s.sref->buf, len);

	sref->nbytes += len;
	sref->buf[sref->nbytes] = 0;

	return *this;
}

AString& AString::operator+=(char s)
{
	// Use temp for G++
	AString tmp(s);
	return *this += tmp;
}

AString& AString::operator+=(const char* _s)
{
	const char *s = (_s ? _s : nil.c_str());

	// Use temp for G++
	AString tmp(s);
	return *this += tmp;
}

AString AString::operator+(char s) const
{
	// Use temp for G++
	AString tmp(*this);
	return tmp += s;
}

AString AString::operator+(const char* s) const
{
	// Use temp for G++
	AString tmp(*this);
	return tmp += s;
}

AString AString::operator+(const AString& s) const
{
	// Use temp for G++
	AString tmp(*this);
	return tmp += s;
}

AString& AString::pad(unsigned len, char ch)
{
	if (sref && (sref->nbytes >= len))
		return *this;

	AStringRef* old_sref = sref;

	sref = new AStringRef;
	sref->buf = new char[len+1];
	sref->refcnt = 1;

	if (old_sref)
	{
		sref->nbytes = old_sref->nbytes;
		memcpy(sref->buf, old_sref->buf, old_sref->nbytes);

		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}
	else
		sref->nbytes = 0;

	while (sref->nbytes < len)
	{
		sref->buf[sref->nbytes++] = ch;
	}

	sref->buf[sref->nbytes] = 0;

	return *this;
}

AString& AString::trim(char ch)
{
	AStringRef* old_sref = sref;
	unsigned len = sref->nbytes;

	while (len > 0)
	{
		if (sref->buf[len-1] != ch)
			break;

		len--;
	}

	sref = new AStringRef;
	sref->buf = new char[len+1];
	sref->refcnt = 1;
	strncpy(sref->buf, old_sref->buf, len);
	sref->buf[len] = 0;
	sref->nbytes = len;

	if (--old_sref->refcnt == 0)
	{
		delete [] old_sref->buf;
		delete old_sref;
	}

	return *this;
}

AString& AString::upper()
{
	if (sref->refcnt > 1)
    {
		AStringRef* old_sref = sref;

		sref = new AStringRef;
		sref->buf = new char[old_sref->nbytes+1];
		sref->refcnt = 1;
		sref->nbytes = old_sref->nbytes;
		memcpy(sref->buf, old_sref->buf, old_sref->nbytes);
		sref->buf[old_sref->nbytes] = 0;

		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}

	char *tmp = sref->buf;

	while (*tmp)
	{
		*tmp = (char)toupper(*tmp);
		tmp++;
	}

	return *this;
}

AString& AString::lower()
{
	if (sref->refcnt > 1)
    {
		AStringRef* old_sref = sref;

		sref = new AStringRef;
		sref->buf = new char[old_sref->nbytes+1];
		sref->refcnt = 1;
		sref->nbytes = old_sref->nbytes;
		memcpy(sref->buf, old_sref->buf, old_sref->nbytes);
		sref->buf[old_sref->nbytes] = 0;

		if (--old_sref->refcnt == 0)
		{
			delete [] old_sref->buf;
			delete old_sref;
		}
	}

	char* tmp = sref->buf;

	while (*tmp)
	{
		*tmp = (char)tolower(*tmp);
		tmp++;
	}

	return *this;
}

AString AString::left(unsigned len) const
{
	return AString(sref->buf + 0, len);
}

AString AString::right(unsigned len) const
{
	unsigned offset;

	if (len < sref->nbytes)
		offset = (unsigned)(sref->nbytes - len);
	else
		offset = 0;

	return AString(sref->buf + offset, len);
}

AString AString::mid(unsigned offset, unsigned len) const
{
	return AString(sref->buf + offset, len);
}


AString AString::dq() const
{
	if ((sref->buf[0] != '\"') && (sref->buf[0] != '\''))
		return *this;

	return AString(sref->buf, (unsigned)1, (unsigned)(length()-2));
}

AString AString::pack() const
{
	AString tmp = "";

	for (unsigned i = 0; i < length(); i++)
	{
		if (  (sref->buf[i] == ' ') ||
				(sref->buf[i] == '\r') ||
				(sref->buf[i] == '\n') )
		{
			tmp += DLE;
			tmp += char(sref->buf[i] + 0x80);
		}
		else
			tmp += sref->buf[i];

	}

	return tmp;
}

AString AString::unpack() const
{
	AString tmp = "";

	for (unsigned i = 0; i < length(); i++)
	{
		if (sref->buf[i] == DLE)
		{
			tmp += char(sref->buf[++i] - 0x80);
		}
		else
			tmp += sref->buf[i];
	}

	return tmp;
}

AString AString::change(char from, char to) const
{
	AString s = data();

	char* tmp = s.sref->buf;

	while (*tmp)
	{
		if (*tmp == from) *tmp = to;
		tmp++;
	}

	return s;
}

AString AString::replace(const char* from, const char* to) const
{
	if (!length() || !from || !to) return data();

	unsigned from_len = strlen(from);
	const char* tmp = data();
	AString s = "";

	while (*tmp)
	{
		if (!strncmp(tmp, from, from_len))
		{
			s += to;
			tmp += from_len;
		}
		else
			s += tmp++;
	}

	return s;
}

AString operator+(char s, const AString& string)
{
	// Use temp for G++
	AString tmp(s);
	return tmp += string;
}

AString operator+(const char* s, const AString& string)
{
	// Use temp for G++
	AString tmp(s);
	return tmp += string;
}

ostream& operator<<(ostream& ostr, const AString& s)
{
	return ostr << s.c_str();
}

istream& operator>>(istream& istr, AString& s)
{
  if (!istr.ipfx(0) || (!(istr.flags() & ios::skipws) && !ws(istr)))
  {
	 istr.clear(ios::failbit|istr.rdstate()); // Redundant if using GNU iostreams.
	 return istr;
  }

  int ch;
  unsigned i = 0;
  const BLOCK_SIZE = 256;
  char* block = (char*)malloc(BLOCK_SIZE);
  unsigned nbytes = BLOCK_SIZE;

  streambuf *sb = istr.rdbuf();
  while ((ch = sb->sbumpc()) != EOF)
  {
	 if (isspace(ch))
		break;

	 if (i == nbytes)
	 {
		nbytes += (unsigned)BLOCK_SIZE;
		block = (char*)realloc(block, nbytes);
	 }

	 block[i++] = (char)ch;
  }

  int new_state = istr.rdstate();
  if (i == 0) new_state |= ios::failbit;
  if (ch == EOF) new_state |= ios::eofbit;
  istr.clear(new_state);

  s = AString(block, 0, i);
  free(block);

  return istr;
}

int AString::Contains(const char* str) const
{
	if (!str || !length()) return 0;

	unsigned len = strlen(str);

	if (!len) return 0;

	for (unsigned i = 0; i < length(); i++)
	{
		if (len > (length() - i))
			break;

		if (!strncmp(sref->buf+i, str, len))
			return 1;
	}

	return 0;
}



