////////////////////////////////////////////////////////////////////////
//
// Description
//
//	 Simple socket class.
//
////////////////////////////////////////////////////////////////////////

#ifndef SIMSOCK_H
#define SIMSOCK_H

////////////////////////////////////////////////////////////////////////

class SimSock
{
public:

	enum SockFlags { NONE=0, NOWAIT=1, FORK=2, ESCAPING=4 };

	virtual ~SimSock();

	virtual int Peek(char* buffer, int nbytes);
	virtual int Read(char* buffer, int nbytes);
	virtual int Write(const char* buffer, int nbytes);
	virtual int Close();

	int OK() { return (fd < 0 ? 0 : 1); };

	virtual int ReadMsg(char* buffer, int nbytes);
	virtual int WriteMsg(const char* buffer, int nbytes, int noframe = 0);

	int IsConnected() const { return connected; };
	int IsDisconnected() const { return !IsConnected(); };
	int IsNoWait() const { return flags & NOWAIT; };
	int IsFork() const { return flags & FORK; };

	// seconds = -1 means indefinite wait
	// seconds = 0 means no wait
	// seconds > 0 means specified wait

	int Wait(long seconds = -1);                   // For read or error
	int WaitForRead(long seconds = -1);
	int WaitForWrite(long seconds = -1);
	int WaitForError(long seconds = -1);

	static const char* LocalHost();

protected:

	enum SockType { SOCK_CLIENT, SOCK_SERVER, SOCK_INTERNAL };

	SimSock(unsigned flags, char eom, SockType type);

	unsigned flags;
	SockType type;
	int connected, fd;

private:

	int _Wait(long seconds, int type);

	static int cnt;			// Keep track of WinSock initialisations
	char eom;

	friend class SockHandler;
	friend class SockServer;
};

////////////////////////////////////////////////////////////////////////

class SockServer : public SimSock
{
public:

	// 'service' can be a name or a port-number

	SockServer(const char* service, unsigned flags = SimSock::NONE, char eom = '\n');

	SimSock* Accept();
};

////////////////////////////////////////////////////////////////////////

class SockClient : public SimSock
{
public:

	SockClient(unsigned flags = SimSock::NONE, char eom = '\n');

	// 'host' can be a name or internet-address-number
	// 'service' can be a name or a port-number

	int Connect(const char* host, const char* service);
	int WaitOnConnect(long seconds = -1);
};

////////////////////////////////////////////////////////////////////////

struct fd_set;
class ListOf;

class SockHandler
{
public:

	SockHandler();
	~SockHandler();

	void Register(SimSock* sock);
	void DeRegister(SimSock* sock);

	unsigned long Count() const;

	// seconds = -1 means indefinite wait
	// seconds = 0 means no wait
	// seconds > 0 means specified wait

	int Wait(long seconds = -1);

	enum SockStatus { SOCK_NONE, SOCK_DATA, SOCK_CONNECT, SOCK_DISCONNECT, SOCK_ERROR };

	SimSock* Notify(SockStatus&);

private:

	ListOf* socks;
	fd_set* read_fds;
	fd_set* write_fds;
	fd_set* except_fds;
};

#endif
